"""
Test `dodecaphony.fragment` module.

Author: Nikolay Lysenko
"""


from collections import Counter
from typing import Any

import pytest

from dodecaphony.fragment import (
    Event,
    Fragment,
    FragmentParams,
    SUPPORTED_DURATIONS,
    calculate_durations_of_measures,
    calculate_number_of_undefined_events,
    create_initial_sonic_content,
    create_initial_temporal_content,
    distribute_pitch_classes,
    find_mutable_sonic_content_indices,
    find_mutable_temporal_content_indices,
    find_sonorities,
    initialize_fragment,
    override_calculated_attributes,
    set_pitches_of_lower_lines,
    set_pitches_of_upper_line,
    split_time_span,
    validate,
)


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A#', 'F#', 'E', 'G#', 'F', 'pause'],
                    ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            [
                [[1.0, 1.0, 1.0, 1.0], [2.0, 2.0], [1.0, 1.0, 1.0, 1.0], [2.0, 1.0, 1.0]],
                [[2.0, 4.0], [2.0, 1.0, 1.0], [1.0, 1.0, 1.0, 1.0], [1.0, 1.0, 1.0, 1.0]],
            ]
        ),
    ]
)
def test_calculate_durations_of_measures(
        fragment: Fragment, expected: list[list[list[float]]]
) -> None:
    """Test `calculate_durations_of_measures` function."""
    fragment = override_calculated_attributes(fragment)
    result = calculate_durations_of_measures(fragment)
    assert result == expected


@pytest.mark.parametrize(
    "group_index, temporal_content, sonic_content, line_indices, n_tone_row_instances, "
    "pauses_fraction, expected",
    [
        (
            # `group_index`
            0,
            # `temporal_content`
            [[], [1.0 for _ in range(12)]],
            # `sonic_content`
            {
                0: {
                    'pitch_classes': [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F',
                    ]
                }
            },
            # `line_indices`
            [0, 1],
            # `n_tone_row_instances`
            3,
            # `pauses_fraction`
            0.0,
            # `expected`
            24
        ),
    ]
)
def test_calculate_number_of_undefined_events(
        group_index: int, temporal_content: list[list[float]],
        sonic_content: dict[int, dict[str, Any]], line_indices: list[int],
        n_tone_row_instances: int, pauses_fraction: float, expected: float
) -> None:
    """Test `calculate_number_of_undefined_events` function."""
    result = calculate_number_of_undefined_events(
        group_index, temporal_content, sonic_content, line_indices, n_tone_row_instances,
        pauses_fraction
    )
    assert result == expected


@pytest.mark.parametrize(
    "params, temporal_content, expected_n_pauses_by_group",
    [
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=100,
                line_ids=[1],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={},
                sonic_content={}
            ),
            # `temporal_content`
            [[1.0, 1.0, 1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]],
            # `expected_n_pauses_by_group`
            [1]
        ),
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=2,
                line_ids=[1],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={},
                sonic_content={
                    0: {
                        'pitch_classes': [
                            'pause', 'B', 'A#', 'G', 'C#', 'D#', 'C',
                            'D', 'A', 'F#', 'E', 'G#', 'F', 'pause'
                        ]
                    }
                }
            ),
            # `temporal_content`
            [[1.0, 1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]],
            # `expected_n_pauses_by_group`
            [2]
        ),
    ]
)
def test_create_initial_sonic_content(
        params: FragmentParams, temporal_content: list[list[float]],
        expected_n_pauses_by_group: list[int]
) -> None:
    """Test `create_initial_sonic_content` function."""
    sonic_content = create_initial_sonic_content(params, temporal_content)
    assert len(sonic_content) == len(params.groups)
    zipped = zip(sonic_content, expected_n_pauses_by_group)
    for i, (line_content, expected_n_pauses) in enumerate(zipped):
        counter = Counter(line_content)
        for pitch_class in params.tone_row:
            assert counter[pitch_class] == params.groups[i]['n_tone_row_instances']
        assert counter['pause'] == expected_n_pauses


@pytest.mark.parametrize(
    "params, expected_n_events_by_line",
    [
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                    {'n_melodic_lines': 3, 'n_tone_row_instances': 6},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2, 3, 4],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={},
                sonic_content={}
            ),
            # `expected_n_events_by_line`
            [13, 27, 27, 26]
        ),
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                    {'n_melodic_lines': 3, 'n_tone_row_instances': 6},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2, 3, 4],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={1: {'durations': [4.0, 4.0, 4.0, 4.0, 4.0, 4.0, 4.0, 4.0]}},
                sonic_content={}
            ),
            # `expected_n_events_by_line`
            [13, 8, 36, 36]
        ),
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                    {'n_melodic_lines': 3, 'n_tone_row_instances': 6},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2, 3, 4],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={},
                sonic_content={
                    0: {
                        'pitch_classes': [
                            'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                        ]
                    }
                }
            ),
            # `expected_n_events_by_line`
            [12, 27, 27, 26]
        ),
    ]
)
def test_create_initial_temporal_content(
        params: FragmentParams, expected_n_events_by_line: list[int]
) -> None:
    """Test `create_initial_temporal_content` function."""
    temporal_content = create_initial_temporal_content(params)
    assert len(temporal_content) == len(params.line_ids)
    n_events_by_line = [len(x) for x in temporal_content]
    assert n_events_by_line == expected_n_events_by_line


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [4.0],
                    [3.0, 1.0],
                    [2.0, 2.0],
                ],
                sonic_content=[
                    ['C'],
                    ['D', 'E', 'F', 'G'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1, 2],
                n_tone_row_instances_by_group=[0, 0],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=4.0, pitch_class='C'),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=3.0, pitch_class='D'),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='G'),
                ],
                [
                    Event(line_index=2, start_time=0.0, duration=2.0, pitch_class='E'),
                    Event(line_index=2, start_time=2.0, duration=2.0, pitch_class='F'),
                ]
            ]
        ),
    ]
)
def test_distribute_pitch_classes(fragment: Fragment, expected: list[list[Event]]) -> None:
    """Test `distribute_pitch_classes` function."""
    result = distribute_pitch_classes(fragment)
    assert result == expected


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                    {'n_melodic_lines': 3, 'n_tone_row_instances': 6},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2, 3, 4],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={},
                sonic_content={
                    0: {
                        'pitch_classes': [
                            'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                        ],
                        'immutable': True
                    }
                }
            ),
            # `expected`
            [1]
        ),
    ]
)
def test_find_mutable_sonic_content_indices(params: FragmentParams, expected: list[int]) -> None:
    """Test `find_mutable_sonic_content_indices` function."""
    result = find_mutable_sonic_content_indices(params)
    assert result == expected


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            # `params`
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                    {'n_melodic_lines': 3, 'n_tone_row_instances': 6},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2, 3, 4],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1,
                temporal_content={
                    1: {
                        'durations': [4.0, 4.0, 4.0, 4.0, 4.0, 4.0, 4.0, 4.0],
                        'immutable': True
                    }
                }
            ),
            # `expected`
            [0, 2, 3]
        ),
    ]
)
def test_find_mutable_temporal_content_indices(
        params: FragmentParams, expected: list[int]
) -> None:
    """Test `find_mutable_temporal_content_indices` function."""
    result = find_mutable_temporal_content_indices(params)
    assert result == expected


@pytest.mark.parametrize(
    "melodic_lines, expected",
    [
        (
            # `melodic_lines`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=3.0),
                    Event(line_index=0, start_time=3.0, duration=1.0),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=2.0),
                    Event(line_index=1, start_time=2.0, duration=2.0),
                ],
                [
                    Event(line_index=2, start_time=0.0, duration=2.0),
                    Event(line_index=2, start_time=2.0, duration=2.0),
                ],
            ],
            # `expected`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=3.0),
                    Event(line_index=1, start_time=0.0, duration=2.0),
                    Event(line_index=2, start_time=0.0, duration=2.0),
                ],
                [
                    Event(line_index=0, start_time=0.0, duration=3.0),
                    Event(line_index=1, start_time=2.0, duration=2.0),
                    Event(line_index=2, start_time=2.0, duration=2.0),
                ],
                [
                    Event(line_index=0, start_time=3.0, duration=1.0),
                    Event(line_index=1, start_time=2.0, duration=2.0),
                    Event(line_index=2, start_time=2.0, duration=2.0),
                ],
            ]
        ),
    ]
)
def test_find_sonorities(melodic_lines: list[list[Event]], expected: list[list[Event]]) -> None:
    """Test `find_sonorities` function."""
    result = find_sonorities(melodic_lines)
    assert result == expected


@pytest.mark.parametrize(
    "params",
    [
        (
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[
                    {'n_melodic_lines': 1, 'n_tone_row_instances': 1},
                    {'n_melodic_lines': 3, 'n_tone_row_instances': 6},
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2, 3, 4],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.1
            )
        ),
    ]
)
def test_initialize_fragment(params: FragmentParams) -> None:
    """Test `initialize_fragment` function."""
    fragment = initialize_fragment(params)
    for melodic_line in fragment.melodic_lines:
        for event in melodic_line:
            assert event.position_in_semitones is not None or event.pitch_class == 'pause'


@pytest.mark.parametrize(
    "fragment, max_interval, default_shift, expected_melodic_lines, expected_sonorities",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['C', 'A', 'D', 'F'],
                    ['D', 'B', 'G', 'A'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[0, 0],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `max_interval`
            16,
            # `default_shift`
            7,
            # `expected_melodic_lines`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B', position_in_semitones=38),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=46),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=36),
                ],
            ],
            # `expected_sonorities`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                ],
                [
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B', position_in_semitones=38),
                ],
                [
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=46),
                ],
                [
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=36),
                ],
            ]
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 1.0, 1.0],
                    [2.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['C', 'D', 'F'],
                    ['C', 'D', 'F'],
                    ['G', 'B', 'G', 'A'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1, 1],
                n_tone_row_instances_by_group=[0, 0, 0],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0, 1, 2],
            ),
            # `max_interval`
            16,
            # `default_shift`
            7,
            # `expected_melodic_lines`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=2.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=2.0, pitch_class='C', position_in_semitones=39),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=32),
                ],
                [
                    Event(line_index=2, start_time=0.0, duration=1.0, pitch_class='G', position_in_semitones=34),
                    Event(line_index=2, start_time=1.0, duration=1.0, pitch_class='B', position_in_semitones=38),
                    Event(line_index=2, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=34),
                    Event(line_index=2, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=24),
                ],
            ],
            # `expected_sonorities`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=2.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=1, start_time=0.0, duration=2.0, pitch_class='C', position_in_semitones=39),
                    Event(line_index=2, start_time=0.0, duration=1.0, pitch_class='G', position_in_semitones=34),
                ],
                [
                    Event(line_index=0, start_time=0.0, duration=2.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=1, start_time=0.0, duration=2.0, pitch_class='C', position_in_semitones=39),
                    Event(line_index=2, start_time=1.0, duration=1.0, pitch_class='B', position_in_semitones=38),
                ],
                [
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                    Event(line_index=2, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=34),
                ],
                [
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=32),
                    Event(line_index=2, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=24),
                ],
            ]
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['C', 'A', 'D', 'F'],
                    ['D', 'pause', 'G', 'A'],
                    ['D', 'B', 'G', 'A'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2, 3],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1, 1],
                n_tone_row_instances_by_group=[0, 0, 0],
                mutable_temporal_content_indices=[0, 1, 2],
                mutable_sonic_content_indices=[0, 1, 2],
            ),
            # `max_interval`
            16,
            # `default_shift`
            24,
            # `expected_melodic_lines`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='pause', position_in_semitones=None),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=46),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=36),
                ],
                [
                    Event(line_index=2, start_time=0.0, duration=1.0, pitch_class='D', position_in_semitones=29),
                    Event(line_index=2, start_time=1.0, duration=1.0, pitch_class='B', position_in_semitones=14),
                    Event(line_index=2, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=22),
                    Event(line_index=2, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=24),
                ],
            ],
            # `expected_sonorities`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                    Event(line_index=2, start_time=0.0, duration=1.0, pitch_class='D', position_in_semitones=29),
                ],
                [
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='pause', position_in_semitones=None),
                    Event(line_index=2, start_time=1.0, duration=1.0, pitch_class='B', position_in_semitones=14),
                ],
                [
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=46),
                    Event(line_index=2, start_time=2.0, duration=1.0, pitch_class='G', position_in_semitones=22),
                ],
                [
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=36),
                    Event(line_index=2, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=24),
                ],
            ]
        ),
    ]
)
def test_set_pitches_of_lower_lines(
        fragment: Fragment,
        max_interval: int,
        default_shift: int,
        expected_melodic_lines: list[list[Event]],
        expected_sonorities: list[list[Event]]
) -> None:
    """Test `set_pitches_of_lower_lines` function."""
    # Below three lines of code are added instead of setting all arguments initially,
    # because `sonorities` and `melodic_lines` must reference to the same events.
    fragment.melodic_lines = distribute_pitch_classes(fragment)
    fragment.sonorities = find_sonorities(fragment.melodic_lines)
    fragment = set_pitches_of_upper_line(fragment)

    fragment = set_pitches_of_lower_lines(fragment, max_interval, default_shift)
    assert fragment.melodic_lines == expected_melodic_lines
    assert fragment.sonorities == expected_sonorities


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['C', 'A', 'D', 'F'],
                    ['D', 'B', 'G', 'A'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[0, 0],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
                melodic_lines=[
                    [
                        Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C'),
                        Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A'),
                        Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D'),
                        Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F'),
                    ],
                    [
                        Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D'),
                        Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B'),
                        Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G'),
                        Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A'),
                    ],
                ],
            ),
            # `expected`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D'),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B'),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G'),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A'),
                ],
            ],
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['pause', 'A', 'D', 'F'],
                    ['D', 'B', 'G', 'A'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[0, 0],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
                melodic_lines=[
                    [
                        Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='pause'),
                        Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A'),
                        Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D'),
                        Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F'),
                    ],
                    [
                        Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D'),
                        Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B'),
                        Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G'),
                        Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A'),
                    ],
                ],
            ),
            # `expected`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='pause', position_in_semitones=None),
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D'),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B'),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G'),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A'),
                ],
            ],
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['C', 'pause', 'D', 'F'],
                    ['D', 'B', 'G', 'A'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[0, 0],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
                melodic_lines=[
                    [
                        Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C'),
                        Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='pause'),
                        Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D'),
                        Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F'),
                    ],
                    [
                        Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D'),
                        Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B'),
                        Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G'),
                        Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A'),
                    ],
                ],
            ),
            # `expected`
            [
                [
                    Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                    Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='pause', position_in_semitones=None),
                    Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                    Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                ],
                [
                    Event(line_index=1, start_time=0.0, duration=1.0, pitch_class='D'),
                    Event(line_index=1, start_time=1.0, duration=1.0, pitch_class='B'),
                    Event(line_index=1, start_time=2.0, duration=1.0, pitch_class='G'),
                    Event(line_index=1, start_time=3.0, duration=1.0, pitch_class='A'),
                ],
            ],
        ),
    ]
)
def test_set_pitches_of_upper_line(fragment: Fragment, expected: list[list[Event]]) -> None:
    """Test `set_pitches_of_upper_line` function."""
    fragment = set_pitches_of_upper_line(fragment)
    assert fragment.melodic_lines == expected


@pytest.mark.parametrize(
    "n_measures, n_events, meter_numerator",
    [
        (2, 9, 4),
        (8, 51, 3),
    ]
)
def test_split_time_span(n_measures: int, n_events: int, meter_numerator: float) -> None:
    """Test `split_time_span` function."""
    durations = split_time_span(n_measures, n_events, meter_numerator)
    assert len(durations) == n_events
    assert sum(durations) == n_measures * meter_numerator
    for duration in durations:
        assert duration in SUPPORTED_DURATIONS


@pytest.mark.parametrize(
    "n_measures, n_events, meter_numerator, match",
    [
        (4, 3, 4, "Average duration of an event is longer than semibreve."),
        (1, 20, 4, "The number of events is so high that some of them are too short.")
    ]
)
def test_split_time_span_with_invalid_arguments(
    n_measures: int, n_events: int, meter_numerator: float, match: str
) -> None:
    """Test `split_time_span` function with invalid arguments."""
    with pytest.raises(ValueError, match=match):
        split_time_span(n_measures, n_events, meter_numerator)


@pytest.mark.parametrize(
    "params, match",
    [
        (
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[{'n_melodic_lines': 1, 'n_tone_row_instances': 2}],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.0
            ),
            "Number of lines in `groups` is not equal to that in `line_ids`."
        ),
        (
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[{'n_melodic_lines': 2, 'n_tone_row_instances': 2}],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 1],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.0
            ),
            "IDs of melodic lines must be unique."
        ),
        (
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[{'n_melodic_lines': 2, 'n_tone_row_instances': 2}],
                meter_numerator=5,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.0
            ),
            "Meter numerator = 5 is not supported."
        ),
        (
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[{'n_melodic_lines': 2, 'n_tone_row_instances': 2}],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.0,
                temporal_content={
                    0: {'durations': [1.0 for _ in range(40)]},
                    1: {'durations': [1.0]},
                }
            ),
            "A line has duration that is not equal to that of the fragment."
        ),
        (
            FragmentParams(
                tone_row=['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                groups=[{'n_melodic_lines': 2, 'n_tone_row_instances': 2}],
                meter_numerator=4,
                meter_denominator=4,
                n_measures=8,
                line_ids=[1, 2],
                upper_line_highest_note='E6',
                upper_line_lowest_note='E4',
                pauses_fraction=0.0,
                sonic_content={
                    0: {
                        'pitch_classes': [
                            'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F',
                            'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F',
                            'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F',
                        ]
                    },
                }
            ),
            "A group has wrong number of tone row instances."
        ),
    ]
)
def test_validate(params: FragmentParams, match: str) -> None:
    """Test `validate` function."""
    with pytest.raises(ValueError, match=match):
        validate(params)


@pytest.mark.parametrize(
    "first_temporal_content, second_temporal_content, first_sonic_content, second_sonic_content, "
    "expected",
    [
        (
            [[1.0 for _ in range(12)]],
            [[1.0 for _ in range(12)]],
            [['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F']],
            [['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F']],
            True
        ),
    ]
)
def test_equality_of_fragments(
        first_temporal_content: list[list[float]], second_temporal_content: list[list[float]],
        first_sonic_content: list[list[str]], second_sonic_content: list[list[str]],
        expected: bool
) -> None:
    """Test `__eq__` method of `Fragment` class."""
    first_fragment = Fragment(
        first_temporal_content,
        first_sonic_content,
        meter_numerator=4,
        meter_denominator=4,
        n_beats=12,
        line_ids=[1],
        upper_line_highest_position=88,
        upper_line_lowest_position=0,
        n_melodic_lines_by_group=[1],
        n_tone_row_instances_by_group=[1],
        mutable_temporal_content_indices=[0],
        mutable_sonic_content_indices=[0]
    )
    first_fragment = override_calculated_attributes(first_fragment)
    second_fragment = Fragment(
        second_temporal_content,
        second_sonic_content,
        meter_numerator=4,
        meter_denominator=4,
        n_beats=12,
        line_ids=[1],
        upper_line_highest_position=88,
        upper_line_lowest_position=0,
        n_melodic_lines_by_group=[1],
        n_tone_row_instances_by_group=[1],
        mutable_temporal_content_indices=[0],
        mutable_sonic_content_indices=[0]
    )
    second_fragment = override_calculated_attributes(second_fragment)
    result = first_fragment == second_fragment
    assert result == expected



"""
Test `dodecaphony.rendering` module.

Author: Nikolay Lysenko
"""


from typing import Any

import pretty_midi
import pytest
import yaml

from dodecaphony.fragment import Fragment, override_calculated_attributes
from dodecaphony.rendering import (
    create_lilypond_file_from_fragment,
    create_midi_from_fragment,
    create_sinethesizer_instruments,
    create_tsv_events_from_fragment,
    create_wav_from_tsv_events,
    create_yaml_from_fragment,
)


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                    [2.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'pause', 'F'],
                    ['F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            (
                "\\version \"2.18.2\"\n"
                "\\layout {\n"
                "    indent = #0\n"
                "}\n"
                "\\new StaffGroup <<\n"
                "    \\new Staff <<\n"
                "        \\clef treble\n"
                "        \\time 4/4\n"
                "        \\key c \\major\n"
                "        {b'4 ais'4 g'4 cis''4 dis''2 c''2 d''4 a'4 fis'4 e'4 gis'2 r4 f'4}\n"
                "    >>\n"
                "    \\new Staff <<\n"
                "        \\clef bass\n"
                "        \\time 4/4\n"
                "        \\key c \\major\n"
                "        {f'2 gis2~ gis2 e'4 fis'4 a'4 d'4 c'4 dis'4 cis'4 g'4 ais4 b4}\n"
                "    >>\n"
                ">>"
            )
        ),
    ]
)
def test_create_lilypond_file_from_fragment(
        path_to_tmp_file: str, fragment: Fragment, expected: str
) -> None:
    """Test `create_lilypond_file_from_fragment` function."""
    fragment = override_calculated_attributes(fragment)
    create_lilypond_file_from_fragment(fragment, path_to_tmp_file)
    with open(path_to_tmp_file) as in_file:
        result = in_file.read()
        assert result == expected


@pytest.mark.parametrize(
    "fragment, note_number, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                    ['F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `note_number`
            5,
            # `expected`
            {
                'pitch': 72,
                'start': 4.0,
                'end': 5.0
            }
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'pause', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                    ['F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `note_number`
            5,
            # `expected`
            {
                'pitch': 74,
                'start': 5.0,
                'end': 5.5
            }
        ),
    ]
)
def test_create_midi_from_fragment(
        path_to_tmp_file: str, fragment: Fragment, note_number: int, expected: dict[str, float]
) -> None:
    """Test `create_midi_from_fragment` function."""
    fragment = override_calculated_attributes(fragment)
    create_midi_from_fragment(
        fragment,
        path_to_tmp_file,
        beat_in_seconds=0.5,
        instruments={k: 0 for k in fragment.line_ids},
        velocity=100,
        opening_silence_in_seconds=1,
        trailing_silence_in_seconds=1
    )
    midi_data = pretty_midi.PrettyMIDI(path_to_tmp_file)
    instrument = midi_data.instruments[0]
    midi_note = instrument.notes[note_number]
    result = {
        'pitch': midi_note.pitch,
        'start': midi_note.start,
        'end': midi_note.end
    }
    assert result == expected


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                    ['F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            (
                "instrument\tstart_time\tduration\tfrequency\tvelocity\teffects\tline_id\n"
                "additive_mellow_pipe\t1.0\t1.0\tF4\t1.0\t\t2\n"
                "additive_mellow_pipe\t1.0\t0.5\tB4\t1.0\t\t1\n"
                "additive_mellow_pipe\t1.5\t0.5\tA#4\t1.0\t\t1\n"
                "additive_mellow_pipe\t2.0\t1.0\tG#3\t1.0\t\t2\n"
                "additive_mellow_pipe\t2.0\t0.5\tG4\t1.0\t\t1\n"
                "additive_mellow_pipe\t2.5\t0.5\tC#5\t1.0\t\t1\n"
                "additive_mellow_pipe\t3.0\t0.5\tE4\t1.0\t\t2\n"
                "additive_mellow_pipe\t3.0\t1.0\tD#5\t1.0\t\t1\n"
                "additive_mellow_pipe\t3.5\t0.5\tF#4\t1.0\t\t2\n"
                "additive_mellow_pipe\t4.0\t0.5\tA4\t1.0\t\t2\n"
                "additive_mellow_pipe\t4.0\t1.0\tC5\t1.0\t\t1\n"
                "additive_mellow_pipe\t4.5\t0.5\tD4\t1.0\t\t2\n"
                "additive_mellow_pipe\t5.0\t1.0\tC4\t1.0\t\t2\n"
                "additive_mellow_pipe\t5.0\t0.5\tD5\t1.0\t\t1\n"
                "additive_mellow_pipe\t5.5\t0.5\tA4\t1.0\t\t1\n"
                "additive_mellow_pipe\t6.0\t1.0\tD#4\t1.0\t\t2\n"
                "additive_mellow_pipe\t6.0\t0.5\tF#4\t1.0\t\t1\n"
                "additive_mellow_pipe\t6.5\t0.5\tE4\t1.0\t\t1\n"
                "additive_mellow_pipe\t7.0\t0.5\tC#4\t1.0\t\t2\n"
                "additive_mellow_pipe\t7.0\t1.0\tG#4\t1.0\t\t1\n"
                "additive_mellow_pipe\t7.5\t0.5\tG4\t1.0\t\t2\n"
                "additive_mellow_pipe\t8.0\t0.5\tA#3\t1.0\t\t2\n"
                "additive_mellow_pipe\t8.0\t1.0\tF4\t1.0\t\t1\n"
                "additive_mellow_pipe\t8.5\t0.5\tB3\t1.0\t\t2\n"
            )
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                    ['pause', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            (
                "instrument\tstart_time\tduration\tfrequency\tvelocity\teffects\tline_id\n"
                "additive_mellow_pipe\t1.0\t0.5\tB4\t1.0\t\t1\n"
                "additive_mellow_pipe\t1.5\t0.5\tA#4\t1.0\t\t1\n"
                "additive_mellow_pipe\t2.0\t1.0\tG#3\t1.0\t\t2\n"
                "additive_mellow_pipe\t2.0\t0.5\tG4\t1.0\t\t1\n"
                "additive_mellow_pipe\t2.5\t0.5\tC#5\t1.0\t\t1\n"
                "additive_mellow_pipe\t3.0\t0.5\tE4\t1.0\t\t2\n"
                "additive_mellow_pipe\t3.0\t1.0\tD#5\t1.0\t\t1\n"
                "additive_mellow_pipe\t3.5\t0.5\tF#4\t1.0\t\t2\n"
                "additive_mellow_pipe\t4.0\t0.5\tA4\t1.0\t\t2\n"
                "additive_mellow_pipe\t4.0\t1.0\tC5\t1.0\t\t1\n"
                "additive_mellow_pipe\t4.5\t0.5\tD4\t1.0\t\t2\n"
                "additive_mellow_pipe\t5.0\t1.0\tC4\t1.0\t\t2\n"
                "additive_mellow_pipe\t5.0\t0.5\tD5\t1.0\t\t1\n"
                "additive_mellow_pipe\t5.5\t0.5\tA4\t1.0\t\t1\n"
                "additive_mellow_pipe\t6.0\t1.0\tD#4\t1.0\t\t2\n"
                "additive_mellow_pipe\t6.0\t0.5\tF#4\t1.0\t\t1\n"
                "additive_mellow_pipe\t6.5\t0.5\tE4\t1.0\t\t1\n"
                "additive_mellow_pipe\t7.0\t0.5\tC#4\t1.0\t\t2\n"
                "additive_mellow_pipe\t7.0\t1.0\tG#4\t1.0\t\t1\n"
                "additive_mellow_pipe\t7.5\t0.5\tG4\t1.0\t\t2\n"
                "additive_mellow_pipe\t8.0\t0.5\tA#3\t1.0\t\t2\n"
                "additive_mellow_pipe\t8.0\t1.0\tF4\t1.0\t\t1\n"
                "additive_mellow_pipe\t8.5\t0.5\tB3\t1.0\t\t2\n"
            )
        ),
    ]
)
def test_create_tsv_events_from_fragment(
        path_to_tmp_file: str, fragment: Fragment, expected: str
) -> None:
    """Test `create_tsv_events_from_fragment` function."""
    fragment = override_calculated_attributes(fragment)
    create_tsv_events_from_fragment(
        fragment,
        path_to_tmp_file,
        beat_in_seconds=0.5,
        instruments={k: 'additive_mellow_pipe' for k in fragment.line_ids},
        effects={k: '' for k in fragment.line_ids},
        velocity=1.0,
        opening_silence_in_seconds=1.0
    )
    with open(path_to_tmp_file) as in_file:
        result = in_file.read()
        assert result == expected


@pytest.mark.parametrize(
    "tsv_content, trailing_silence_in_seconds",
    [
        (
            [
                'instrument\tstart_time\tduration\tfrequency\tvelocity\teffects\tline_id',
                'additive_mellow_pipe\t1\t1\tA0\t1\t\t1',
                'additive_mellow_pipe\t2\t1\t1\t1\t[{"name": "tremolo", "frequency": 1}]\t1'
            ],
            1.0
        ),
    ]
)
def test_create_wav_from_tsv_events(
        path_to_tmp_file: str, path_to_another_tmp_file: str,
        tsv_content: list[str], trailing_silence_in_seconds: float
) -> None:
    """Test `create_wav_from_tsv_events` function."""
    with open(path_to_tmp_file, 'w') as tmp_tsv_file:
        for line in tsv_content:
            tmp_tsv_file.write(line + '\n')
    instruments_registry = create_sinethesizer_instruments()
    create_wav_from_tsv_events(
        path_to_tmp_file,
        path_to_another_tmp_file,
        instruments_registry,
        trailing_silence_in_seconds
    )


@pytest.mark.parametrize(
    "fragment, expected",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0],
                    [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                    ['pause', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=16,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[1, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected`
            {
                "temporal_content": {
                    0: {'durations': [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0]},
                    1: {'durations': [2.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 1.0, 1.0, 1.0, 1.0]},
                },
                'sonic_content': {
                    0: {'pitch_classes': ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F']},
                    1: {'pitch_classes': ['pause', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B']},
                }
            }
        ),
    ]
)
def test_create_yaml_from_fragment(
        path_to_tmp_file: str, fragment: Fragment, expected: dict[str, Any]
) -> None:
    """Test `create_yaml_from_fragment` function."""
    create_yaml_from_fragment(fragment, path_to_tmp_file)
    with open(path_to_tmp_file) as in_file:
        result = yaml.load(in_file, Loader=yaml.FullLoader)
    assert result == expected



"""
Do auxiliary tasks.

Author: Nikolay Lysenko
"""


import multiprocessing as mp
from typing import Any, Callable, Optional


def starmap_in_parallel(
        fn: Callable, args: list[Any], pool_kwargs: Optional[dict[str, Any]] = None
) -> list[Any]:
    """
    Apply function to each collection of arguments from the given list in parallel.

    This function contains boilerplate code that is needed for correct work
    of `pytest-cov`. Usage of `mp.Pool` as context manager is not alternative
    to this function, because:
    1) not all covered lines of code may be marked as covered
       (see more: https://pytest-cov.readthedocs.io/en/latest/subprocess-support.html);
    2) some files with names like '.coverage.hostname.*' may be not deleted
       (see more: https://github.com/pytest-dev/pytest-cov/issues/250).

    :param fn:
        function
    :param args:
        list of collections of arguments
    :param pool_kwargs:
        parameters of pool such as number of processes and maximum number of
        tasks for a worker before it is replaced with a new one
    :return:
        results of applying the function to the arguments
    """
    pool_kwargs = pool_kwargs or {}
    key_renaming = {'n_processes': 'processes', 'max_tasks_per_child': 'maxtasksperchild'}
    pool_kwargs = {key_renaming.get(k, k): v for k, v in pool_kwargs.items()}
    pool = mp.Pool(**pool_kwargs)
    try:
        results = pool.starmap(fn, args)
    finally:
        pool.close()
        pool.join()
    return results


def compute_rolling_aggregate(
        values: list[float], aggregation_fn: Callable[[list[float]], float], window_size: int
) -> list[float]:
    """
    Compute rolling aggregate.

    :param values:
        list of values to be aggregated
    :param aggregation_fn:
        aggregation function
    :param window_size:
        size of rolling window
    :return:
        list of rolling aggregates
    """
    window = []
    results = []
    for value in values:
        if len(window) == window_size:
            window.pop(0)
        window.append(value)
        results.append(aggregation_fn(window))
    return results



"""
Iteratively improve fragment.

Author: Nikolay Lysenko
"""


import copy
import math
import os
from dataclasses import dataclass
from typing import Any, Optional

from .evaluation import SCORING_SETS_REGISTRY_TYPE, evaluate
from .fragment import Fragment
from .transformations import TRANSFORMATION_REGISTRY_TYPE, transform
from .utils import starmap_in_parallel


@dataclass
class Record:
    """An evaluated fragment and its score."""
    fragment: Fragment
    score: float


@dataclass
class Task:
    """Task for a single process."""
    incumbent_solution: Fragment
    n_trials: int


def select_distinct_best_records(records: list[Record], n_records: int) -> list[Record]:
    """
    Select records related to highest scores (without duplicates).

    :param records:
        original records
    :param n_records:
        number of unique records to select
    :return:
        best records
    """
    results = []
    for record in sorted(records, key=lambda x: -x.score):
        if record not in results:
            results.append(record)
        if len(results) == n_records:
            break
    return results


def generate_new_records(
        tasks: list[Task],
        n_records_to_return: int,
        n_transformations_per_trial: int,
        transformation_registry: TRANSFORMATION_REGISTRY_TYPE,
        transformation_names: list[str],
        transformation_probabilities: list[float],
        scoring_sets: list[str],
        scoring_sets_registry: SCORING_SETS_REGISTRY_TYPE
) -> list[Record]:  # pragma: no cover  # Actually, it is covered, but `pytest-cov` misses it.
    """
    Generate records for given tasks.

    :param tasks:
        fragments such that their neighborhoods should be searched and numbers of transformed
        fragments to generate and evaluate for each of them
    :param n_records_to_return:
        number of best records to return; this argument is added to the function for the sake of
        performance, because it reduces amount of data returned by a child process to a
        parent process
    :param n_transformations_per_trial:
        number of transformations to be applied to generate a new fragment
    :param transformation_registry:
        mapping from names to corresponding transformations and their arguments
    :param transformation_names:
        names of transformations to choose from
    :param transformation_probabilities:
        probabilities of corresponding transformations; this argument must have the same length
        as `transformation_names`
    :param scoring_sets:
        names of scoring sets to be used for fragments evaluation
    :param scoring_sets_registry:
        mapping from a name of a scoring set to a list of triples of a scoring function,
        its weight, and its parameters
    :return:
        best generated records
    """
    new_records = []
    for task in tasks:
        incumbent_solution = task.incumbent_solution
        for trial_id in range(task.n_trials):
            candidate = Fragment(
                [copy.copy(x) for x in incumbent_solution.temporal_content],
                [copy.copy(x) for x in incumbent_solution.sonic_content],
                incumbent_solution.meter_numerator,
                incumbent_solution.meter_denominator,
                incumbent_solution.n_beats,
                incumbent_solution.line_ids,
                incumbent_solution.upper_line_highest_position,
                incumbent_solution.upper_line_lowest_position,
                incumbent_solution.n_melodic_lines_by_group,
                incumbent_solution.n_tone_row_instances_by_group,
                incumbent_solution.mutable_temporal_content_indices,
                incumbent_solution.mutable_sonic_content_indices
            )
            candidate = transform(
                candidate,
                n_transformations_per_trial,
                transformation_registry,
                transformation_names,
                transformation_probabilities
            )
            score, _ = evaluate(candidate, scoring_sets, scoring_sets_registry)
            score = round(score, 10)  # Prevent overflow during records comparison.
            new_records.append(Record(candidate, score))
    new_records = select_distinct_best_records(new_records, n_records_to_return)
    return new_records


def create_tasks(
        incumbent_solutions: list[Fragment],
        n_trials_per_iteration: int,
        paralleling_params: dict[str, Any]
) -> list[list[Task]]:
    """
    Distribute incumbent solutions and trials to improve them between processes.

    :param incumbent_solutions:
        fragments such that their neighborhoods should be searched
    :param n_trials_per_iteration:
        number of transformed fragments to generate and evaluate per each incumbent solution
    :param paralleling_params:
        settings of parallel running of trials
    :return:
        tasks for each process
    """
    raw_tasks = []
    for incumbent_solution in incumbent_solutions:
        raw_tasks.append(Task(incumbent_solution, n_trials_per_iteration))
    n_trials = len(incumbent_solutions) * n_trials_per_iteration
    n_processes = paralleling_params.get('n_processes') or os.cpu_count()
    n_trials_per_process = math.ceil(n_trials / n_processes)
    remaining_n_trials_per_current_process = n_trials_per_process
    tasks = [[Task(incumbent_solutions[0], 0)]]
    i = 0
    while i < len(raw_tasks):
        if raw_tasks[i].incumbent_solution not in [task.incumbent_solution for task in tasks[-1]]:
            tasks[-1].append(Task(raw_tasks[i].incumbent_solution, 0))
        n_trials_to_add = min(remaining_n_trials_per_current_process, raw_tasks[i].n_trials)
        tasks[-1][-1].n_trials += n_trials_to_add
        remaining_n_trials_per_current_process -= n_trials_to_add
        raw_tasks[i].n_trials -= n_trials_to_add
        if remaining_n_trials_per_current_process == 0:
            tasks.append([])
            remaining_n_trials_per_current_process = n_trials_per_process
        if raw_tasks[i].n_trials == 0:
            i += 1
    if not tasks[-1]:
        tasks.pop()
    return tasks


def optimize_with_local_search(
        fragment: Fragment,
        n_iterations: int,
        n_trials_per_iteration: int,
        default_n_transformations_per_trial: int,
        n_transformations_increment: int,
        max_n_transformations_per_trial: int,
        beam_width: int,
        transformation_registry: TRANSFORMATION_REGISTRY_TYPE,
        transformation_probabilities: dict[str, float],
        scoring_sets: list[str],
        scoring_sets_registry: SCORING_SETS_REGISTRY_TYPE,
        paralleling_params: Optional[dict[str, Any]] = None
) -> list[Fragment]:
    """
    Optimize initial fragment with local beam search.

    :param fragment:
        initial fragment
    :param n_iterations:
        number of optimization steps where a step is generation of new incumbent solutions
        by transformation of previous incumbent solutions
    :param n_trials_per_iteration:
        number of transformed fragments to generate and evaluate per each incumbent solution
        at each iteration
    :param default_n_transformations_per_trial:
        default number of transformations to be applied to generate a new fragment
    :param n_transformations_increment:
        increment of number of transformations added after every iteration such that global
        best score has not been improved after it; default number is used again as soon as
        global best score improves
    :param max_n_transformations_per_trial:
        maximum number of transformations to be applied to generate a new fragment
    :param beam_width:
        number of incumbent solutions to consider at each iteration
    :param transformation_registry:
        mapping from names to corresponding transformations and their arguments
    :param transformation_probabilities:
        mapping from transformation name to its probability
    :param scoring_sets:
        names of scoring sets to be used for fragments evaluation
    :param scoring_sets_registry:
        mapping from a name of a scoring set to a list of triples of a scoring function,
        its weight, and its parameters
    :param paralleling_params:
        settings of parallel running of trials
    :return:
        optimized fragments
    """
    incumbent_solutions = [fragment]
    records = []
    previous_best_score = -1e9
    n_transformations_per_trial = default_n_transformations_per_trial
    transformation_names = list(transformation_probabilities.keys())
    transformation_probabilities = list(transformation_probabilities.values())
    paralleling_params = paralleling_params or {}

    for iteration_id in range(n_iterations):
        all_tasks = create_tasks(incumbent_solutions, n_trials_per_iteration, paralleling_params)
        args = [
            (
                tasks_for_process,
                beam_width,
                n_transformations_per_trial,
                transformation_registry,
                transformation_names,
                transformation_probabilities,
                scoring_sets,
                scoring_sets_registry
            )
            for tasks_for_process in all_tasks
        ]
        nested_new_records = starmap_in_parallel(generate_new_records, args, paralleling_params)
        new_records = [record for records in nested_new_records for record in records]
        best_new_records = select_distinct_best_records(new_records, beam_width)
        current_best_score = best_new_records[0].score
        incumbent_solutions = [record.fragment for record in best_new_records]

        if current_best_score > previous_best_score:
            n_transformations_per_trial = default_n_transformations_per_trial
        # If optimization stucks, Iterated Local Search suggests to perturb the current solutions.
        # Here, increase of neighborhood size is a form of perturbation.
        else:
            n_transformations_per_trial += n_transformations_increment
        # If enough perturbations are made, let us switch back to regular neighborhoods.
        if n_transformations_per_trial > max_n_transformations_per_trial:
            n_transformations_per_trial = default_n_transformations_per_trial

        records = select_distinct_best_records(records + new_records, beam_width)
        global_best_score = records[0].score
        previous_best_score = current_best_score
        print(
            f'Iteration #{iteration_id:>3}: '
            f'global_best_score = {global_best_score:.5f}, '
            f'current_best_score = {current_best_score:.5f}'
        )
    result = [record.fragment for record in records]
    return result



"""
Test `dodecaphony.music_theory` module.

Author: Nikolay Lysenko
"""


from typing import Optional

import pytest

from dodecaphony.music_theory import (
    get_mapping_from_pitch_class_to_diatonic_scales,
    get_smallest_intervals_between_pitch_classes,
    invert_tone_row,
    revert_tone_row,
    rotate_tone_row,
    transpose_tone_row,
    validate_tone_row,
)


@pytest.mark.parametrize(
    "scale_types, key, value",
    [
        (
            # `scale_types`
            None,
            # `key`
            'C',
            # `value`
            [
                'C-major',
                'C#-major',
                'D#-major',
                'F-major',
                'G-major',
                'G#-major',
                'A#-major',
                'C-natural_minor',
                'D-natural_minor',
                'E-natural_minor',
                'F-natural_minor',
                'G-natural_minor',
                'A-natural_minor',
                'A#-natural_minor',
                'C-harmonic_minor',
                'C#-harmonic_minor',
                'E-harmonic_minor',
                'F-harmonic_minor',
                'G-harmonic_minor',
                'A-harmonic_minor',
                'A#-harmonic_minor',
                'C-dorian',
                'D-dorian',
                'D#-dorian',
                'F-dorian',
                'G-dorian',
                'A-dorian',
                'A#-dorian',
                'C-phrygian',
                'D-phrygian',
                'E-phrygian',
                'F-phrygian',
                'G-phrygian',
                'A-phrygian',
                'B-phrygian',
                'C-lydian',
                'C#-lydian',
                'D#-lydian',
                'F-lydian',
                'F#-lydian',
                'G#-lydian',
                'A#-lydian',
                'C-mixolydian',
                'D-mixolydian',
                'D#-mixolydian',
                'F-mixolydian',
                'G-mixolydian',
                'G#-mixolydian',
                'A#-mixolydian',
                'C-locrian',
                'D-locrian',
                'E-locrian',
                'F#-locrian',
                'G-locrian',
                'A-locrian',
                'B-locrian',
                'C-whole_tone',
            ]
        ),
        (
            # `scale_types`
            ('major', 'whole_tone'),
            # `key`
            'C#',
            # `value`
            [
                'C#-major',
                'D-major',
                'E-major',
                'F#-major',
                'G#-major',
                'A-major',
                'B-major',
                'C#-whole_tone',
            ]
        ),
    ]
)
def test_get_mapping_from_pitch_class_to_diatonic_scales(
        scale_types: Optional[list[str]], key: str, value: list[str]
) -> None:
    """Test `get_mapping_from_pitch_class_to_diatonic_scales` function."""
    mapping = get_mapping_from_pitch_class_to_diatonic_scales(scale_types)
    assert mapping[key] == value


@pytest.mark.parametrize(
    "key, value",
    [
        (('C', 'B'), -1),
        (('C', 'F#'), 6),
    ]
)
def test_get_smallest_intervals_between_pitch_classes(key: tuple[str, str], value: int) -> None:
    """Test `get_smallest_intervals_between_pitch_classes` function."""
    mapping = get_smallest_intervals_between_pitch_classes()
    assert mapping[key] == value


@pytest.mark.parametrize(
    "tone_row, expected",
    [
        (
            ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
            ['B', 'C', 'D#', 'A', 'G', 'A#', 'G#', 'C#', 'E', 'F#', 'D', 'F'],
        ),
    ]
)
def test_invert_tone_row(tone_row: list[str], expected: list[str]) -> None:
    """Test `invert_tone_row` function."""
    result = invert_tone_row(tone_row)
    assert result == expected


@pytest.mark.parametrize(
    "tone_row, expected",
    [
        (
            ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
            ['F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'],
        ),
    ]
)
def test_revert_tone_row(tone_row: list[str], expected: list[str]) -> None:
    """Test `revert_tone_row` function."""
    result = revert_tone_row(tone_row)
    assert result == expected


@pytest.mark.parametrize(
    "tone_row, shift, expected",
    [
        (
            ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
            -1,
            ['A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'B'],
        ),
        (
            ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
            3,
            ['E', 'G#', 'F', 'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#'],
        ),
    ]
)
def test_rotate_tone_row(tone_row: list[str], shift: int, expected: list[str]) -> None:
    """Test `rotate_tone_row` function."""
    result = rotate_tone_row(tone_row, shift)
    assert result == expected


@pytest.mark.parametrize(
    "tone_row, shift_in_semitones, expected",
    [
        (
            ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
            -1,
            ['A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'],
        ),
    ]
)
def test_transpose_tone_row(
        tone_row: list[str], shift_in_semitones: int, expected: list[str]
) -> None:
    """Test `transpose_tone_row` function."""
    result = transpose_tone_row(tone_row, shift_in_semitones)
    assert result == expected


@pytest.mark.parametrize(
    "tone_row, match",
    [
        (
            ['A', 'B'],
            "Tone row must have 12 elements."
        ),
        (
            ['B', 'A', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
            "All pitch classes must be included in a tone row."
        ),
    ]
)
def test_validate_tone_row(tone_row: list[str], match: str) -> None:
    """Test `validate_tone_row` function."""
    with pytest.raises(ValueError, match=match):
        validate_tone_row(tone_row)



"""
Provide utilities for working with pitches, pitch classes, and tone rows.

Author: Nikolay Lysenko
"""


import itertools
from enum import Enum
from functools import cache
from typing import Optional


N_SEMITONES_PER_OCTAVE = 12
TONE_ROW_LEN = 12
PITCH_CLASS_TO_POSITION = {
    'C': 0, 'C#': 1, 'D': 2, 'D#': 3, 'E': 4, 'F': 5,
    'F#': 6, 'G': 7, 'G#': 8, 'A': 9, 'A#': 10, 'B': 11
}
POSITION_TO_PITCH_CLASS = {v: k for k, v in PITCH_CLASS_TO_POSITION.items()}


class IntervalTypes(Enum):
    """Enumeration of interval types."""
    PERFECT_CONSONANCE = 1
    IMPERFECT_CONSONANCE = 2
    DISSONANCE = 3
    NOT_AN_INTERVAL = 4


N_SEMITONES_TO_INTERVAL_TYPE_WITH_CONSONANT_P4 = {
    0: IntervalTypes.PERFECT_CONSONANCE,
    1: IntervalTypes.DISSONANCE,
    2: IntervalTypes.DISSONANCE,
    3: IntervalTypes.IMPERFECT_CONSONANCE,
    4: IntervalTypes.IMPERFECT_CONSONANCE,
    5: IntervalTypes.IMPERFECT_CONSONANCE,
    6: IntervalTypes.DISSONANCE,
    7: IntervalTypes.PERFECT_CONSONANCE,
    8: IntervalTypes.IMPERFECT_CONSONANCE,
    9: IntervalTypes.IMPERFECT_CONSONANCE,
    10: IntervalTypes.DISSONANCE,
    11: IntervalTypes.DISSONANCE,
}
N_SEMITONES_TO_INTERVAL_TYPE_WITH_DISSONANT_P4 = {
    0: IntervalTypes.PERFECT_CONSONANCE,
    1: IntervalTypes.DISSONANCE,
    2: IntervalTypes.DISSONANCE,
    3: IntervalTypes.IMPERFECT_CONSONANCE,
    4: IntervalTypes.IMPERFECT_CONSONANCE,
    5: IntervalTypes.DISSONANCE,
    6: IntervalTypes.DISSONANCE,
    7: IntervalTypes.PERFECT_CONSONANCE,
    8: IntervalTypes.IMPERFECT_CONSONANCE,
    9: IntervalTypes.IMPERFECT_CONSONANCE,
    10: IntervalTypes.DISSONANCE,
    11: IntervalTypes.DISSONANCE,
}


@cache
def get_smallest_intervals_between_pitch_classes() -> dict[tuple[str, str], int]:
    """
    Get mapping from a pair of pitch classes to the smallest interval connecting them.

    Tritone intervals are stored as upward intervals.

    :return:
        mapping from a pair of pitch classes to the smallest interval (in semitones)
        connecting them
    """
    pitch_classes = PITCH_CLASS_TO_POSITION.keys()
    cartesian_product = itertools.product(pitch_classes, pitch_classes)
    result = {}
    for starting_pitch_class, destination_pitch_class in cartesian_product:
        destination_position = PITCH_CLASS_TO_POSITION[destination_pitch_class]
        starting_position = PITCH_CLASS_TO_POSITION[starting_pitch_class]
        shift = N_SEMITONES_PER_OCTAVE // 2 - 1
        value = (destination_position - starting_position + shift) % N_SEMITONES_PER_OCTAVE - shift
        result[(starting_pitch_class, destination_pitch_class)] = value
    return result


def get_type_of_interval(
        n_semitones: int, is_perfect_fourth_consonant: bool = True
) -> IntervalTypes:
    """
    Get type of a harmonic interval.

    :param n_semitones:
        interval size in semitones
    :param is_perfect_fourth_consonant:
        indicator whether perfect fourth is a consonant interval
    :return:
        type of interval
    """
    if is_perfect_fourth_consonant:
        n_semitones_to_consonance = N_SEMITONES_TO_INTERVAL_TYPE_WITH_CONSONANT_P4
    else:
        n_semitones_to_consonance = N_SEMITONES_TO_INTERVAL_TYPE_WITH_DISSONANT_P4
    n_semitones %= len(n_semitones_to_consonance)
    return n_semitones_to_consonance[n_semitones]


def validate_tone_row(tone_row: list[str]) -> None:
    """
    Validate tone row.

    :param tone_row:
        tone row as list of pitch classes (like C or C#, flats are not allowed)
    :return:
        None
    """
    if len(tone_row) != TONE_ROW_LEN:
        raise ValueError("Tone row must have 12 elements.")
    positions = [PITCH_CLASS_TO_POSITION[pitch_class] for pitch_class in tone_row]
    if sorted(positions) != list(range(TONE_ROW_LEN)):
        raise ValueError("All pitch classes must be included in a tone row.")


def invert_tone_row(tone_row: list[str]) -> list[str]:
    """
    Invert tone row preserving its first pitch class.

    :param tone_row:
        tone row as list of pitch classes (like C or C#, flats are not allowed)
    :return:
        inverted tone row
    """
    inverted_tone_row = [tone_row[0]]
    for former_pitch_class, latter_pitch_class in zip(tone_row, tone_row[1:]):
        latter_position = PITCH_CLASS_TO_POSITION[latter_pitch_class]
        former_position = PITCH_CLASS_TO_POSITION[former_pitch_class]
        interval = latter_position - former_position
        next_position = (PITCH_CLASS_TO_POSITION[inverted_tone_row[-1]] - interval)
        next_position %= N_SEMITONES_PER_OCTAVE
        next_pitch_class = POSITION_TO_PITCH_CLASS[next_position]
        inverted_tone_row.append(next_pitch_class)
    return inverted_tone_row


def revert_tone_row(tone_row: list[str]) -> list[str]:
    """
    Revert tone row, i.e., apply retrograde inversion to it.

    :param tone_row:
        tone row as list of pitch classes (like C or C#, flats are not allowed)
    :return:
        reverted tone row
    """
    return tone_row[::-1]


def rotate_tone_row(tone_row: list[str], shift: int) -> list[str]:
    """
    Rotate tone row.

    :param tone_row:
        tone row as list of pitch classes (like C or C#, flats are not allowed)
    :param shift:
        shift; if positive; tone row is rotated forward, if negative, tone row is rotated backwards
    :return:
        rotated tone row
    """
    return tone_row[-shift:] + tone_row[:-shift]


def transpose_tone_row(tone_row: list[str], shift_in_semitones: int) -> list[str]:
    """
    Transpose tone row.

    :param tone_row:
        tone row as list of pitch classes (like C or C#, flats are not allowed)
    :param shift_in_semitones:
        transposition interval in semitones
    :return:
        transposed tone row
    """
    transposed_tone_row = []
    for pitch_class in tone_row:
        new_position = (PITCH_CLASS_TO_POSITION[pitch_class] + shift_in_semitones)
        new_position %= N_SEMITONES_PER_OCTAVE
        new_pitch_class = POSITION_TO_PITCH_CLASS[new_position]
        transposed_tone_row.append(new_pitch_class)
    return transposed_tone_row


@cache
def get_mapping_from_pitch_class_to_diatonic_scales(
        scale_types: Optional[tuple[str]] = None
) -> dict[str, list[str]]:
    """
    Get mapping from pitch class to list of names of diatonic scales to which it belongs.

    :param scale_types:
        types of diatonic scales to be tested; this list may include the following values:
        'major', 'natural_minor', 'harmonic_minor', 'dorian', 'phrygian', 'lydian', 'mixolydian',
        'locrian', and 'whole_tone'
    :return:
        mapping from pitch class to list of names of diatonic scales to which it belongs
    """
    patterns = {
        'major': [1, 0, 2, 0, 3, 4, 0, 5, 0, 6, 0, 7],
        'natural_minor': [1, 0, 2, 3, 0, 4, 0, 5, 6, 0, 7, 0],
        'harmonic_minor': [1, 0, 2, 3, 0, 4, 0, 5, 6, 0, 0, 7],
        'dorian': [1, 0, 2, 3, 0, 4, 0, 5, 0, 6, 7, 0],
        'phrygian': [1, 2, 0, 3, 0, 4, 0, 5, 6, 0, 7, 0],
        'lydian': [1, 0, 2, 0, 3, 0, 4, 5, 0, 6, 0, 7],
        'mixolydian': [1, 0, 2, 0, 3, 4, 0, 5, 0, 6, 7, 0],
        'locrian': [1, 2, 0, 3, 0, 4, 5, 0, 6, 0, 7, 0],
        'whole_tone': [1, 0, 2, 0, 3, 0, 4, 0, 5, 0, 6, 0],
    }
    if scale_types is not None:
        patterns = {k: v for k, v in patterns.items() if k in scale_types}
    pitch_classes = list(PITCH_CLASS_TO_POSITION.keys())
    result = {pitch_class: [] for pitch_class in pitch_classes}
    cartesian_product = itertools.product(patterns.items(), enumerate(pitch_classes))
    for (scale_type, pattern), (offset, pitch_class) in cartesian_product:
        if scale_type == 'whole_tone' and pitch_class not in ['C', 'C#']:
            continue  # Prevent creation of duplicated scales.
        scale_name = f'{pitch_class}-{scale_type}'
        rotated_pitch_classes = pitch_classes[offset:] + pitch_classes[:offset]
        for another_pitch_class, degree in zip(rotated_pitch_classes, pattern):
            if degree:
                result[another_pitch_class].append(scale_name)
    return result



"""
Render fragment to some formats such as WAV and MIDI.

Author: Nikolay Lysenko
"""


import datetime
import os
import subprocess
import traceback
from math import ceil, floor
from pkg_resources import resource_filename
from typing import Any

import pretty_midi
import yaml
from sinethesizer.io import (
    convert_events_to_timeline,
    convert_tsv_to_events,
    create_instruments_registry,
    write_timeline_to_wav
)
from sinethesizer.synth.core import Instrument
from sinethesizer.utils.music_theory import get_list_of_notes, get_note_to_position_mapping

from .fragment import Event, Fragment


NOTE_TO_POSITION = get_note_to_position_mapping()
POSITION_TO_NOTE = {v: k for k, v in NOTE_TO_POSITION.items()}


def create_midi_from_fragment(
        fragment: Fragment,
        midi_path: str,
        beat_in_seconds: float,
        instruments: dict[int, int],
        velocity: int,
        opening_silence_in_seconds: float = 1.0,
        trailing_silence_in_seconds: float = 1.0
) -> None:
    """
    Create MIDI file from a fragment created by this package.

    :param fragment:
        musical fragment
    :param midi_path:
        path where resulting MIDI file is going to be saved
    :param beat_in_seconds:
        duration of one beat in seconds
    :param instruments:
        mapping from IDs of melodic lines to IDs of instruments (according to General MIDI
        specification) that play them
    :param velocity:
        one common velocity for all notes
    :param opening_silence_in_seconds:
        number of seconds with silence to add at the start of the composition
    :param trailing_silence_in_seconds:
        number of seconds with silence to add at the end of the composition
    :return:
        None
    """
    numeration_shift = pretty_midi.note_name_to_number('A0')
    pretty_midi_instruments = []
    for line_id, melodic_line in zip(fragment.line_ids, fragment.melodic_lines):
        pretty_midi_instrument = pretty_midi.Instrument(instruments[line_id], name=str(line_id))
        for event in melodic_line:
            if event.pitch_class in ['pause', 'skip']:
                continue
            start_time = event.start_time * beat_in_seconds
            start_time += opening_silence_in_seconds
            end_time = start_time + event.duration * beat_in_seconds
            pitch = event.position_in_semitones + numeration_shift
            note = pretty_midi.Note(
                start=start_time,
                end=end_time,
                pitch=pitch,
                velocity=velocity
            )
            pretty_midi_instrument.notes.append(note)
        pretty_midi_instrument.notes.sort(key=lambda x: (x.start, x.pitch))
        pretty_midi_instruments.append(pretty_midi_instrument)

    trailing_silence_start = fragment.n_beats * beat_in_seconds
    trailing_silence_start += opening_silence_in_seconds
    note = pretty_midi.Note(
        start=trailing_silence_start,
        end=trailing_silence_start + trailing_silence_in_seconds,
        velocity=0,
        pitch=1  # Arbitrary value that affects nothing.
    )
    pretty_midi_instruments[0].notes.append(note)

    composition = pretty_midi.PrettyMIDI()
    for pretty_midi_instrument in pretty_midi_instruments:
        composition.instruments.append(pretty_midi_instrument)
    composition.write(midi_path)


def create_tsv_events_from_fragment(
        fragment: Fragment,
        events_path: str,
        beat_in_seconds: float,
        instruments: dict[int, str],
        effects: dict[int, str],
        velocity: float,
        opening_silence_in_seconds: float = 1.0
) -> None:
    """
    Create TSV file with `sinethesizer` events from a fragment.

    :param fragment:
        musical fragment
    :param events_path:
        path to a file where result is going to be saved
    :param beat_in_seconds:
        duration of one beat in seconds
    :param instruments:
        mapping from IDs of melodic lines to names of instruments (from a `sinethesizer` preset)
        that play them
    :param effects:
        mapping from IDs of melodic lines to sound effects to be applied to all events
        from the corresponding lines
    :param velocity:
        one common velocity for all events
    :param opening_silence_in_seconds:
        number of seconds with silence to add at the start of the composition
    :return:
        None
    """
    all_notes = get_list_of_notes()
    events = []
    for line_id, melodic_line in zip(fragment.line_ids, fragment.melodic_lines):
        instrument = instruments[line_id]
        line_effects = effects.get(line_id, '')
        for event in melodic_line:
            if event.pitch_class in ['pause', 'skip']:
                continue
            start_time = event.start_time * beat_in_seconds
            start_time += opening_silence_in_seconds
            duration_in_seconds = event.duration * beat_in_seconds
            pitch_id = event.position_in_semitones
            note = all_notes[pitch_id]
            event = (
                instrument,
                start_time,
                duration_in_seconds,
                note,
                pitch_id,
                line_effects,
                line_id
            )
            events.append(event)
    events = sorted(events, key=lambda x: (x[1], x[4], x[2]))
    events = [f"{x[0]}\t{x[1]}\t{x[2]}\t{x[3]}\t{velocity}\t{x[5]}\t{x[6]}" for x in events]

    columns = [
        'instrument',
        'start_time',
        'duration',
        'frequency',
        'velocity',
        'effects',
        'line_id'
    ]
    header = '\t'.join(columns)
    results = [header] + events
    with open(events_path, 'w') as out_file:
        for line in results:
            out_file.write(line + '\n')


def create_sinethesizer_instruments() -> dict[str, Instrument]:
    """
    Create registry of `sinethesizer` instruments.

    :return:
        mapping from instrument names to instruments itself
    """
    presets_path = resource_filename('dodecaphony', 'configs/sinethesizer_presets.yml')
    instruments_registry = create_instruments_registry(presets_path)
    return instruments_registry


def create_wav_from_tsv_events(
        events_path: str, output_path: str,
        instruments_registry: dict[str, Instrument],
        trailing_silence_in_seconds: float
) -> None:
    """
    Create WAV file based on `sinethesizer` TSV file.

    :param events_path:
        path to TSV file with track represented as `sinethesizer` events
    :param output_path:
        path where resulting WAV file is going to be saved
    :param instruments_registry:
        mapping from instrument names to instruments itself
    :param trailing_silence_in_seconds:
        number of seconds with silence to add at the end of the composition
    :return:
        None
    """
    settings = {
        'frame_rate': 48000,
        'trailing_silence': trailing_silence_in_seconds,
        'peak_amplitude': 1,
        'instruments_registry': instruments_registry,
    }
    events = convert_tsv_to_events(events_path, settings)
    timeline = convert_events_to_timeline(events, settings)
    write_timeline_to_wav(output_path, timeline, settings['frame_rate'])


def create_yaml_from_fragment(fragment: Fragment, yaml_path: str) -> None:
    """
    Create YAML file that can be used for setting temporal and sonic content in a runtime config.

    :param fragment:
        musical fragment
    :param yaml_path:
        path to a file where result is going to be saved
    :return:
        None
    """
    temporal_content = {
        i: {'durations': durations}
        for i, durations in enumerate(fragment.temporal_content)
    }
    sonic_content = {
        i: {'pitch_classes': pitch_classes}
        for i, pitch_classes in enumerate(fragment.sonic_content)
    }
    result = {'temporal_content': temporal_content, 'sonic_content': sonic_content}
    with open(yaml_path, 'w') as out_file:
        yaml.dump(result, out_file, default_flow_style=None)


def make_lilypond_template(
        n_voices: int,
        tonic: str,
        scale_type: str,
        meter_numerator: int,
        meter_denominator: int
) -> str:
    """
    Make template of Lilypond text file.

    :param n_voices:
        number of voices in a fragment to be rendered
    :param tonic:
        tonic pitch class represented by letter (like C or A#)
    :param scale_type:
        type of scale (e.g., 'major', 'natural_minor', 'harmonic_minor', 'dorian', and so on)
    :param meter_numerator:
        numerator in meter signature, i.e., number of reference beats per measure
    :param meter_denominator:
        denominator in meter signature, i.e., ratio of reference beat duration to the whole note
    :return:
        template
    """
    raw_template = (
        "\\version \"2.18.2\"\n"
        "\\layout {{{{\n"
        "    indent = #0\n"
        "}}}}\n"
        "\\new StaffGroup <<\n"
        "    \\new Staff <<\n"
        "        \\clef treble\n"
        "        \\time {}/{}\n"
        "        \\key {} \\{}\n"
        "{}"
        "    >>\n"
        "    \\new Staff <<\n"
        "        \\clef bass\n"
        "        \\time {}/{}\n"
        "        \\key {} \\{}\n"
        "{}"
        "    >>\n"
        ">>"
    )
    tonic = tonic.replace('#', 'is').replace('b', 'es').lower()
    scale_type = scale_type.split('_')[-1]
    voices = ["        {{{}}}\n" for _ in range(n_voices)]
    treble_bass_threshold = ceil(n_voices / 2)
    template = raw_template.format(
        meter_numerator, meter_denominator, tonic, scale_type,
        "        \\\\\n".join(voices[:treble_bass_threshold]),
        meter_numerator, meter_denominator, tonic, scale_type,
        "        \\\\\n".join(voices[treble_bass_threshold:])
    )
    return template


def get_lilypond_order_of_voices(n_voices: int) -> list[int]:
    """
    Enumerate voices (from highest to lowest) in Lilypond order.

    See more about Lilypond order ('Voice order' section):
    http://lilypond.org/doc/v2.18/Documentation/notation/multiple-voices

    :param n_voices:
        number of voices in a fragment to be rendered
    :return:
        indices of voices in Lilypond order
    """
    def enumerate_for_one_staff(n_voices_at_staff: int) -> list[int]:
        max_index = n_voices_at_staff - 1
        results = []
        for i in range(n_voices_at_staff):
            result = max_index - int(round(2 * abs(i - max_index / 2)))
            result += int(i < max_index / 2)
            results.append(result)
        return results

    n_voices_at_upper_staff = ceil(n_voices / 2)
    n_voices_at_lower_staff = floor(n_voices / 2)
    lower_voices_priorities = [
        x + n_voices_at_upper_staff
        for x in enumerate_for_one_staff(n_voices_at_lower_staff)
    ]
    upper_voices_priorities = enumerate_for_one_staff(n_voices_at_upper_staff)
    priorities = lower_voices_priorities + upper_voices_priorities
    ordering = sorted(
        list((index, priority) for index, priority in enumerate(priorities)),
        key=lambda x: x[1], reverse=True
    )
    ordering = [x[0] for x in ordering]
    return ordering


def find_lilypond_duration(
        duration: float,
        time_in_measure: float,
        meter_numerator: int,
        meter_denominator: int
) -> list[str]:
    """
    Find duration of a note in Lilypond reciprocal format.

    :param duration:
        duration of a note in beats
    :param time_in_measure:
        number of previous beats passed from the start of the current measure
    :param meter_numerator:
        numerator in meter signature, i.e., number of reference beats per measure
    :param meter_denominator:
        denominator in meter signature, i.e., ratio of reference beat duration to the whole note
    :return:
        strings representing Lilypond durations of notes (a note can be split to multiple notes
        if it crosses bar or if its duration is compound and requires ties)
    """
    if time_in_measure + duration <= meter_numerator:
        reciprocal_duration = meter_denominator / duration
        supported_reciprocal_durations = {
            16: ['16'],
            32 / 3: ['16.'],
            8: ['8'],
            16 / 3: ['8.'],
            4: ['4'],
            3.2: ['4~', '16'],
            1 / 0.34375: ['4~', '16.'],
            8 / 3: ['4.'],
            1 / 0.4375: ['4.~', '16'],
            1 / 0.46875: ['4.~', '16.'],
            2: ['2'],
            1 / 0.5625: ['2~', '16'],
            1 / 0.59375: ['2~', '16.'],
            1.6: ['2~', '8'],
            1 / 0.6875: ['2~', '8.'],
            4 / 3: ['2.'],
            1 / 0.8125: ['2.~', '16'],
            1 / 0.84375: ['2.~', '16.'],
            8 / 7: ['2.~', '8'],
            1 / 0.9375: ['2.~', '8.'],
            1: ['1'],
            2 / 3: ['1.'],
            0.5: ['\\breve'],
            1 / 3: ['\\breve.'],
            0.25: ['\\longa'],
            1 / 6: ['\\longa.'],
        }
        lilypond_duration = supported_reciprocal_durations.get(reciprocal_duration)
        if lilypond_duration is None:
            raise RuntimeError(f"Reciprocal duration {reciprocal_duration} is not supported yet.")
        return lilypond_duration
    else:
        remaining_duration = meter_numerator - time_in_measure
        remaining_results = find_lilypond_duration(
            remaining_duration,
            time_in_measure,
            meter_numerator,
            meter_denominator
        )
        remaining_results[-1] = remaining_results[-1] + '~'
        left_over_bar_duration = duration - meter_numerator + time_in_measure
        left_over_bar_results = find_lilypond_duration(
            left_over_bar_duration,
            0,
            meter_numerator,
            meter_denominator
        )
        return remaining_results + left_over_bar_results


def convert_to_lilypond_note(
        event: Event,
        meter_numerator: int,
        meter_denominator: int
) -> str:
    """
    Convert `Event` instance to note in Lilypond absolute notation.

    :param event:
        event
    :param meter_numerator:
        numerator in meter signature, i.e., number of reference beats per measure
    :param meter_denominator:
        denominator in meter signature, i.e., ratio of reference beat duration to the whole note
    :return:
        note in Lilypond absolute notation
    """
    if event.pitch_class == 'pause':
        note_without_duration = 'r'
    else:
        pitch_class = event.pitch_class
        pitch_class = pitch_class.replace('#', 'is').replace('b', 'es')
        pitch_class = pitch_class.lower()

        octave_id = int(POSITION_TO_NOTE[event.position_in_semitones][-1])
        lilypond_default_octave_id = 3
        octave_diff = octave_id - lilypond_default_octave_id
        octave_sign = "'" if octave_diff >= 0 else ','
        octave_info = "".join(octave_sign for _ in range(abs(octave_diff)))

        note_without_duration = pitch_class + octave_info

    start_time = event.start_time
    time_in_measure = start_time % meter_numerator
    durations = find_lilypond_duration(
        event.duration,
        time_in_measure,
        meter_numerator,
        meter_denominator
    )
    note = [f"{note_without_duration}{duration}" for duration in durations]
    note = " ".join(note)
    if event.pitch_class in ['pause', 'skip']:
        note = note.replace('~', '')  # Lilypond warns when pauses or skips are tied.
    return note


def create_lilypond_file_from_fragment(fragment: Fragment, output_path: str) -> None:
    """
    Create text file in format of Lilypond sheet music editor.

    :param fragment:
        musical fragment
    :param output_path:
        path where resulting file is going to be saved
    :return:
        None
    """
    n_voices = len(fragment.melodic_lines)
    template = make_lilypond_template(
        n_voices,
        # For a dodecaphonic piece, there is no tonic and no scale type
        # and the two below constants affect nothing.
        'C',
        'major',
        fragment.meter_numerator,
        fragment.meter_denominator
    )
    lilypond_voices = []
    indices = get_lilypond_order_of_voices(n_voices)
    for index in indices:
        melodic_line = fragment.melodic_lines[index]
        lilypond_voice = []
        for event in melodic_line:
            note = convert_to_lilypond_note(event, fragment.meter_numerator, fragment.meter_denominator)
            lilypond_voice.append(note)
        lilypond_voice = " ".join(lilypond_voice)
        lilypond_voices.append(lilypond_voice)
    result = template.format(*lilypond_voices)
    with open(output_path, 'w') as out_file:
        out_file.write(result)


def create_pdf_sheet_music_with_lilypond(lilypond_path: str) -> None:  # pragma: no cover
    """
    Create PDF file with sheet music.

    :param lilypond_path:
        path to a text file in Lilypond format
    :return:
        None:
    """
    dir_path, filename = os.path.split(lilypond_path)
    bash_command = f"lilypond {filename}"
    try:
        process = subprocess.Popen(
            bash_command.split(),
            cwd=dir_path,
            stdout=subprocess.PIPE
        )
        process.communicate()
    except Exception:
        print("Rendering sheet music to PDF failed. Do you have Lilypond?")
        print(traceback.format_exc())


def render(fragment: Fragment, rendering_params: dict[str, Any]) -> None:  # pragma: no cover
    """
    Save fragment to MIDI, WAV, TSV, PDF, and Lilypond files.

    :param fragment:
        musical fragment
    :param rendering_params:
        settings of fragment saving
    :return:
        None
    """
    top_level_dir = rendering_params['dir']
    now = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S,%f")
    nested_dir = os.path.join(top_level_dir, f"result_{now}")
    os.mkdir(nested_dir)

    midi_path = os.path.join(nested_dir, 'music.mid')
    midi_params = rendering_params['midi']
    common_params = rendering_params['common'].copy()
    create_midi_from_fragment(fragment, midi_path, **midi_params, **common_params)

    events_path = os.path.join(nested_dir, 'sinethesizer_events.tsv')
    events_params = rendering_params['sinethesizer']
    trailing_silence_in_sec = common_params.pop('trailing_silence_in_seconds')
    create_tsv_events_from_fragment(fragment, events_path, **events_params, **common_params)

    wav_path = os.path.join(nested_dir, 'music.wav')
    instruments_registry = create_sinethesizer_instruments()
    create_wav_from_tsv_events(events_path, wav_path, instruments_registry, trailing_silence_in_sec)

    yaml_path = os.path.join(nested_dir, 'content.yml')
    create_yaml_from_fragment(fragment, yaml_path)

    meta_information_path = os.path.join(nested_dir, 'meta_information.txt')
    with open(meta_information_path, 'w') as in_file:
        in_file.write(rendering_params['meta_information'])

    lilypond_path = os.path.join(nested_dir, 'sheet_music.ly')
    create_lilypond_file_from_fragment(fragment, lilypond_path)
    create_pdf_sheet_music_with_lilypond(lilypond_path)



"""
Test `dodecaphony.optimization` module.

Author: Nikolay Lysenko
"""


from typing import Any

import pytest

from dodecaphony.evaluation import parse_scoring_sets_registry
from dodecaphony.fragment import Fragment
from dodecaphony.optimization import (
    Record, Task, create_tasks, optimize_with_local_search, select_distinct_best_records
)
from dodecaphony.transformations import create_transformations_registry


@pytest.mark.parametrize(
    "incumbent_solutions, n_trials_per_iteration, paralleling_params, expected",
    [
        (
            # `incumbent_solutions`
            [1, 2, 3, 4, 5],
            # `n_trials_per_iteration`
            1000,
            # `paralleling_params`
            {'n_processes': 8},
            # `expected`
            [
                [Task(1, 625)],
                [Task(1, 375), Task(2, 250)],
                [Task(2, 625)],
                [Task(2, 125), Task(3, 500)],
                [Task(3, 500), Task(4, 125)],
                [Task(4, 625)],
                [Task(4, 250), Task(5, 375)],
                [Task(5, 625)],
            ]
        ),
        (
            # `incumbent_solutions`
            [1, 2, 3, 4, 5],
            # `n_trials_per_iteration`
            1000,
            # `paralleling_params`
            {'n_processes': 6},
            # `expected`
            [
                [Task(1, 834)],
                [Task(1, 166), Task(2, 668)],
                [Task(2, 332), Task(3, 502)],
                [Task(3, 498), Task(4, 336)],
                [Task(4, 664), Task(5, 170)],
                [Task(5, 830)],
            ]
        ),
    ]
)
def test_create_tasks(
        incumbent_solutions: list[int], n_trials_per_iteration: int,
        paralleling_params: dict[str, Any], expected: list[list[Task]]
) -> None:
    """Test `create_tasks` function."""
    result = create_tasks(incumbent_solutions, n_trials_per_iteration, paralleling_params)
    assert result == expected


@pytest.mark.parametrize(
    "fragment, n_iterations, n_trials_per_iteration, default_n_transformations_per_trial, "
    "n_transformations_increment, max_n_transformations_per_trial, beam_width, max_rotation, "
    "max_transposition, transformation_probabilities, scoring_sets, scoring_sets_params",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1, 2],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[2, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `n_iterations`
            10,
            # `n_trials_per_iteration`
            10,
            # `default_n_transformations_per_trial`
            1,
            # `n_transformations_increment`
            1,
            # `max_n_transformations_per_trial`
            2,
            # `beam_width`
            1,
            # `max_rotation`
            1,
            # `max_transposition`
            1,
            # `transformation_probabilities`
            {
                'duration_change': 0.25,
                'inversion': 0.25,
                'reversion': 0.25,
                'transposition': 0.25,
            },
            # `scoring_sets`
            ['default'],
            # `scoring_sets_params`
            [
                {
                    'name': 'default',
                    'scoring_functions': [
                        {
                            'name': 'smoothness_of_voice_leading',
                            'weights': {0.0: 1.0},
                            'penalty_deduction_per_line': 0.2,
                            'n_semitones_to_penalty': {
                                0: 0.2,
                                1: 0.0,
                                2: 0.0,
                                3: 0.1,
                                4: 0.2,
                                5: 0.3,
                                6: 0.4,
                                7: 0.5,
                                8: 0.6,
                                9: 0.7,
                                10: 0.8,
                                11: 0.9,
                                12: 1.0,
                            },
                        },
                    ],
                }
            ],
        ),
    ]
)
def test_optimize_with_local_search(
        fragment: Fragment, n_iterations: int, n_trials_per_iteration: int,
        default_n_transformations_per_trial: int, n_transformations_increment: int,
        max_n_transformations_per_trial: int, beam_width: int, max_rotation: int,
        max_transposition: int, transformation_probabilities: dict[str, float],
        scoring_sets: list[str], scoring_sets_params: list[dict[str, Any]]
) -> None:
    """Test `optimize_with_local_search` function."""
    transformations_registry = create_transformations_registry(max_rotation, max_transposition)
    scoring_sets_registry = parse_scoring_sets_registry(scoring_sets_params)
    optimize_with_local_search(
        fragment, n_iterations, n_trials_per_iteration, default_n_transformations_per_trial,
        n_transformations_increment, max_n_transformations_per_trial, beam_width,
        transformations_registry, transformation_probabilities, scoring_sets, scoring_sets_registry
    )


@pytest.mark.parametrize(
    "records, n_records, expected",
    [
        ([], 5, []),
        ([Record('a', 0), Record('b', -1), Record('a', 0)], 2, [Record('a', 0), Record('b', -1)]),
    ]
)
def test_select_distinct_best_records(
        records: list[Record], n_records: int, expected: list[Record]
) -> None:
    """Test `select_distinct_best_records` function."""
    result = select_distinct_best_records(records, n_records)
    assert result == expected



"""
Test `dodecaphony.transformations` module.

Author: Nikolay Lysenko
"""


import pytest

from dodecaphony.fragment import Event, Fragment
from dodecaphony.transformations import (
    apply_duration_change,
    apply_inversion,
    apply_pause_shift,
    apply_reversion,
    apply_rotation,
    apply_transposition,
    create_transformations_registry,
    get_duration_changes,
    transform,
)


@pytest.mark.parametrize(
    "fragment, expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 2.0],
                    [2.0, 1.0, 1.0],
                ],
                sonic_content=[
                    ['A', 'B', 'C', 'D', 'E']
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=4,
                line_ids=[1, 2],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[2],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0],
            ),
            # `expected_options`
            [
                [
                    [2.0, 2.0],
                    [2.0, 1.0, 1.0],
                ],
                [
                    [1.0, 3.0],
                    [2.0, 1.0, 1.0],
                ],
                [
                    [3.0, 1.0],
                    [2.0, 1.0, 1.0],
                ],
                [
                    [2.0, 2.0],
                    [1.0, 2.0, 1.0],
                ],
                [
                    [2.0, 2.0],
                    [1.0, 1.0, 2.0],
                ],
                [
                    [2.0, 2.0],
                    [1.5, 1.5, 1.0],
                ],
                [
                    [2.0, 2.0],
                    [1.5, 1.0, 1.5],
                ],
                [
                    [2.0, 2.0],
                    [2.0, 0.5, 1.5],
                ],
                [
                    [2.0, 2.0],
                    [2.0, 1.5, 0.5],
                ],
            ]
        )
    ]
)
def test_apply_duration_change(
        fragment: Fragment, expected_options: list[list[list[list[Event]]]]
) -> None:
    """Test `apply_duration_change` function."""
    fragment = apply_duration_change(fragment, get_duration_changes())
    assert fragment.temporal_content in expected_options


@pytest.mark.parametrize(
    "fragment, expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1, 2],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[2, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected_options`
            [
                [
                    [
                        'B', 'C', 'D#', 'A', 'G', 'A#', 'G#', 'C#', 'E', 'F#', 'D', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'C', 'D#', 'A', 'G', 'A#', 'G#', 'C#', 'E', 'F#', 'D', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'C', 'D#', 'A', 'G', 'A#', 'G#', 'C#', 'E', 'F#', 'D', 'F'
                    ],
                ],
            ]
        ),
    ]
)
def test_apply_inversion(fragment: Fragment, expected_options: list[list[list[str]]]) -> None:
    """Test `apply_inversion` function."""
    fragment = apply_inversion(fragment)
    assert fragment.sonic_content in expected_options


@pytest.mark.parametrize(
    "fragment, expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                ],
                sonic_content=[
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1],
                n_tone_row_instances_by_group=[2],
                mutable_temporal_content_indices=[0],
                mutable_sonic_content_indices=[0],
            ),
            # `expected_options`
            [
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'pause', 'F',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'B',
                        'pause', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
            ]
        ),
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0],
                mutable_sonic_content_indices=[0],
            ),
            # `expected_options`
            [
                [
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                ],
            ]
        ),
    ]
)
def test_apply_pause_shift(fragment: Fragment, expected_options: list[list[list[str]]]) -> None:
    """Test `apply_pause_shift` function."""
    fragment = apply_pause_shift(fragment)
    assert fragment.sonic_content in expected_options


@pytest.mark.parametrize(
    "fragment, expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1, 2],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[2, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `expected_options`
            [
                [
                    [
                        'F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'F', 'G#', 'E', 'F#', 'A', 'D', 'C', 'D#', 'C#', 'G', 'A#', 'B'
                    ],
                ],
            ]
        ),
    ]
)
def test_apply_reversion(fragment: Fragment, expected_options: list[list[list[str]]]) -> None:
    """Test `apply_reversion` function."""
    fragment = apply_reversion(fragment)
    assert fragment.sonic_content in expected_options


@pytest.mark.parametrize(
    "fragment, max_rotation, expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1, 2],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[2, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `max_rotation`
            1,
            # `expected_options`
            [
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'B', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'F', 'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'B'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'F', 'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'B'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'F', 'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#'
                    ],
                ],
            ]
        ),
    ]
)
def test_apply_rotation(
        fragment: Fragment, max_rotation: int, expected_options: list[list[list[str]]]
) -> None:
    """Test `apply_rotation` function."""
    fragment = apply_rotation(fragment, max_rotation)
    assert fragment.sonic_content in expected_options


@pytest.mark.parametrize(
    "fragment, max_transposition, expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                    [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                ],
                sonic_content=[
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=24,
                line_ids=[1, 2],
                upper_line_highest_position=88,
                upper_line_lowest_position=1,
                n_melodic_lines_by_group=[1, 1],
                n_tone_row_instances_by_group=[2, 1],
                mutable_temporal_content_indices=[0, 1],
                mutable_sonic_content_indices=[0, 1],
            ),
            # `max_transposition`
            1,
            # `expected_options`
            [
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'A#', 'A', 'F#', 'C', 'D', 'B', 'C#', 'G#', 'F', 'D#', 'G', 'E'
                    ],
                ],
                [
                    [
                        'C', 'B', 'G#', 'D', 'E', 'C#', 'D#', 'A#', 'G', 'F', 'A', 'F#', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'C', 'B', 'G#', 'D', 'E', 'C#', 'D#', 'A#', 'G', 'F', 'A', 'F#'
                    ],
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                ],
                [
                    [
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F', 'pause',
                        'B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'
                    ],
                    [
                        'C', 'B', 'G#', 'D', 'E', 'C#', 'D#', 'A#', 'G', 'F', 'A', 'F#'
                    ],
                ],
            ]
        ),
    ]
)
def test_apply_transposition(
        fragment: Fragment, max_transposition: int, expected_options: list[list[list[str]]]
) -> None:
    """Test `apply_transposition` function."""
    fragment = apply_transposition(fragment, max_transposition)
    assert fragment.sonic_content in expected_options


@pytest.mark.parametrize(
    "key, value",
    [
        ((0.5, 2.0), [(0.5, 2.0), (1.0, 1.5), (1.5, 1.0), (2.0, 0.5)]),
    ]
)
def test_get_duration_changes(key: tuple[float, float], value: list[tuple[float, float]]) -> None:
    """Test `get_duration_changes` function."""
    assert get_duration_changes()[key] == value


@pytest.mark.parametrize(
    "fragment, n_transformations, transformation_names, max_rotation, max_transposition, "
    "expected_options",
    [
        (
            # `fragment`
            Fragment(
                temporal_content=[
                    [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
                ],
                sonic_content=[
                    ['B', 'A#', 'G', 'C#', 'D#', 'C', 'D', 'A', 'F#', 'E', 'G#', 'F'],
                ],
                meter_numerator=4,
                meter_denominator=4,
                n_beats=12,
                line_ids=[1],
                upper_line_highest_position=55,
                upper_line_lowest_position=41,
                n_melodic_lines_by_group=[1],
                n_tone_row_instances_by_group=[1],
                mutable_temporal_content_indices=[0],
                mutable_sonic_content_indices=[0],
            ),
            # `n_transformations`
            1,
            # `transformation_names`
            ['inversion', 'reversion'],
            # `max_rotation`
            1,
            # `max_transposition`
            1,
            # `expected_options`
            [
                [
                    [
                        Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='B', position_in_semitones=50),
                        Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                        Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='D#', position_in_semitones=54),
                        Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                        Event(line_index=0, start_time=4.0, duration=1.0, pitch_class='G', position_in_semitones=46),
                        Event(line_index=0, start_time=5.0, duration=1.0, pitch_class='A#', position_in_semitones=49),
                        Event(line_index=0, start_time=6.0, duration=1.0, pitch_class='G#', position_in_semitones=47),
                        Event(line_index=0, start_time=7.0, duration=1.0, pitch_class='C#', position_in_semitones=52),
                        Event(line_index=0, start_time=8.0, duration=1.0, pitch_class='E', position_in_semitones=55),
                        Event(line_index=0, start_time=9.0, duration=1.0, pitch_class='F#', position_in_semitones=45),
                        Event(line_index=0, start_time=10.0, duration=1.0, pitch_class='D', position_in_semitones=41),
                        Event(line_index=0, start_time=11.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                    ]
                ],
                [
                    [
                        Event(line_index=0, start_time=0.0, duration=1.0, pitch_class='F', position_in_semitones=44),
                        Event(line_index=0, start_time=1.0, duration=1.0, pitch_class='G#', position_in_semitones=47),
                        Event(line_index=0, start_time=2.0, duration=1.0, pitch_class='E', position_in_semitones=43),
                        Event(line_index=0, start_time=3.0, duration=1.0, pitch_class='F#', position_in_semitones=45),
                        Event(line_index=0, start_time=4.0, duration=1.0, pitch_class='A', position_in_semitones=48),
                        Event(line_index=0, start_time=5.0, duration=1.0, pitch_class='D', position_in_semitones=53),
                        Event(line_index=0, start_time=6.0, duration=1.0, pitch_class='C', position_in_semitones=51),
                        Event(line_index=0, start_time=7.0, duration=1.0, pitch_class='D#', position_in_semitones=54),
                        Event(line_index=0, start_time=8.0, duration=1.0, pitch_class='C#', position_in_semitones=52),
                        Event(line_index=0, start_time=9.0, duration=1.0, pitch_class='G', position_in_semitones=46),
                        Event(line_index=0, start_time=10.0, duration=1.0, pitch_class='A#', position_in_semitones=49),
                        Event(line_index=0, start_time=11.0, duration=1.0, pitch_class='B', position_in_semitones=50),
                    ]
                ],
            ]
        ),
    ]
)
def test_transform(
        fragment: Fragment, n_transformations: int, transformation_names: list[str],
        max_rotation: int, max_transposition: int, expected_options: list[list[list[Event]]]
) -> None:
    """Test `transform` function."""
    registry = create_transformations_registry(max_rotation, max_transposition)
    transformation_probabilities = [1 / len(transformation_names) for _ in transformation_names]
    fragment = transform(
        fragment, n_transformations, registry, transformation_names, transformation_probabilities
    )
    assert fragment.melodic_lines in expected_options



"""
Transform a fragment in progress.

Note that intermediate functions from this module modify only `temporal_content` and
`sonic_content` attributes, but `melodic_lines` and `sonorities` attributes are left unchanged.
This is done for the sake of performance. It is cheaper to update all dependent attributes
just once after all transformations are applied. So use `transform` function to get
consistent fragment.

Author: Nikolay Lysenko
"""


import itertools
import random
from typing import Any, Callable

from .fragment import Fragment, SUPPORTED_DURATIONS, override_calculated_attributes
from .music_theory import (
    TONE_ROW_LEN, invert_tone_row, revert_tone_row, rotate_tone_row, transpose_tone_row
)


TRANSFORMATION_REGISTRY_TYPE = dict[str, tuple[Callable, list[Any]]]


def get_duration_changes() -> dict[tuple[float, float], list[tuple[float, float]]]:
    """
    Get mapping from durations of two events to list of pairs of durations of the same total sum.

    :return:
        mapping from durations of two events to list of pairs of durations of the same total sum
    """
    result = {}
    cartesian_product = itertools.product(SUPPORTED_DURATIONS, SUPPORTED_DURATIONS)
    for first_duration, second_duration in cartesian_product:
        if first_duration > second_duration:
            continue
        total_sum = first_duration + second_duration
        durations = []
        for duration in SUPPORTED_DURATIONS:
            complementary_duration = total_sum - duration
            if complementary_duration in SUPPORTED_DURATIONS:
                durations.append((duration, complementary_duration))
        result[(first_duration, second_duration)] = durations
    return result


def draw_random_indices(
        mutable_sonic_content_indices: list[int], n_tone_row_instances_by_group: list[int]
) -> tuple[int, int]:
    """
    Draw index of melodic lines group and index of tone row instance from it.

    :param mutable_sonic_content_indices:
        indices of groups such that their sonic content can be transformed
    :param n_tone_row_instances_by_group:
        list where number of tone row instances in the group is stored for each group of
        melodic lines sharing the same series (in terms of vertical distribution of series pitches)
    :return:
        index of group and index of tone row instance from it
    """
    group_index = random.choice(mutable_sonic_content_indices)
    n_instances = n_tone_row_instances_by_group[group_index]
    instance_index = random.randrange(0, n_instances)
    return group_index, instance_index


def find_instance_by_indices(
        fragment: Fragment, group_index: int, instance_index: int
) -> list[str]:
    """
    Find sequence of 12 pitch classes by its indices.

    :param fragment:
        a fragment
    :param group_index:
        index of group of melodic lines sharing the same series (in terms of vertical distribution
        of series pitches)
    :param instance_index:
        index of tone row instance within sonic content of the group
    :return:
        sequence of 12 pitch classes
    """
    sequence = fragment.sonic_content[group_index]
    start_event_index = instance_index * TONE_ROW_LEN
    end_event_index = (instance_index + 1) * TONE_ROW_LEN
    pitch_classes = []
    index = 0
    for pitch_class in sequence:
        if pitch_class == 'pause':
            continue
        if index >= end_event_index:
            break
        if index >= start_event_index:
            pitch_classes.append(pitch_class)
        index += 1
    return pitch_classes


def replace_instance(
        fragment: Fragment, group_index: int, instance_index: int, new_instance: list[str]
) -> Fragment:
    """
    Replace a particular sequence of 12 tones with another sequence of 12 tones.

    :param fragment:
        a fragment
    :param group_index:
        index of group of melodic lines sharing the same series (in terms of vertical distribution
        of series pitches)
    :param instance_index:
        index of tone row instance within sonic content of the group
    :param new_instance:
        new sequence of 12 pitch classes
    :return:
        modified fragment
    """
    sequence = fragment.sonic_content[group_index]
    start_event_index = instance_index * TONE_ROW_LEN
    end_event_index = (instance_index + 1) * TONE_ROW_LEN
    index_without_pauses = 0
    for index, pitch_class in enumerate(sequence):
        if pitch_class == 'pause':
            continue
        if index_without_pauses >= end_event_index:
            break
        if index_without_pauses >= start_event_index:
            sequence[index] = new_instance.pop(0)
        index_without_pauses += 1
    return fragment


def apply_duration_change(
        fragment: Fragment,
        duration_changes: dict[tuple[float, float], list[tuple[float, float]]]
) -> Fragment:
    """
    Change durations of two random events from the same melodic line.

    :param fragment:
        a fragment to be modified
    :param duration_changes:
        mapping from durations of two events to list of pairs of durations of the same total sum
    :return:
        modified fragment
    """
    line_index = random.choice(fragment.mutable_temporal_content_indices)
    line_durations = fragment.temporal_content[line_index]
    events_indices = random.sample(range(len(line_durations)), 2)
    key = tuple(sorted(line_durations[event_index] for event_index in events_indices))
    all_durations = duration_changes[key]
    durations = random.choice(all_durations)
    for event_index, duration in zip(events_indices, durations):
        line_durations[event_index] = duration
    return fragment


def apply_pause_shift(fragment: Fragment) -> Fragment:
    """
    Shift a random pause one position to the left or to the right.

    :param fragment:
        a fragment to be modified
    :return:
        modified fragment
    """
    group_index = random.choice(fragment.mutable_sonic_content_indices)
    sequence = fragment.sonic_content[group_index]
    indices = []
    for index, (previous_pitch_class, pitch_class) in enumerate(zip(sequence, sequence[1:])):
        if pitch_class == 'pause' and previous_pitch_class != 'pause':
            indices.append(index)
        if pitch_class != 'pause' and previous_pitch_class == 'pause':
            indices.append(index)
    if not indices:
        return fragment
    index = random.choice(indices)
    sequence[index], sequence[index + 1] = sequence[index + 1], sequence[index]
    return fragment


def apply_inversion(fragment: Fragment) -> Fragment:
    """
    Invert one random series (transformed tone row instance).

    :param fragment:
        a fragment to be modified
    :return:
        modified fragment
    """
    group_index, instance_index = draw_random_indices(
        fragment.mutable_sonic_content_indices, fragment.n_tone_row_instances_by_group
    )
    tone_row_instance = find_instance_by_indices(fragment, group_index, instance_index)
    tone_row_instance = invert_tone_row(tone_row_instance)
    fragment = replace_instance(fragment, group_index, instance_index, tone_row_instance)
    return fragment


def apply_reversion(fragment: Fragment) -> Fragment:
    """
    Revert one random series (transformed tone row instance).

    :param fragment:
        a fragment to be modified
    :return:
        modified fragment
    """
    group_index, instance_index = draw_random_indices(
        fragment.mutable_sonic_content_indices, fragment.n_tone_row_instances_by_group
    )
    tone_row_instance = find_instance_by_indices(fragment, group_index, instance_index)
    tone_row_instance = revert_tone_row(tone_row_instance)
    fragment = replace_instance(fragment, group_index, instance_index, tone_row_instance)
    return fragment


def apply_rotation(fragment: Fragment, max_rotation: int) -> Fragment:
    """
    Rotate one random series (transformed tone row instance).

    :param fragment:
        a fragment to be modified
    :param max_rotation:
        maximum size of rotation (in elements)
    :return:
        modified fragment
    """
    group_index, instance_index = draw_random_indices(
        fragment.mutable_sonic_content_indices, fragment.n_tone_row_instances_by_group
    )
    tone_row_instance = find_instance_by_indices(fragment, group_index, instance_index)
    shift = random.randint(-max_rotation, max_rotation)
    tone_row_instance = rotate_tone_row(tone_row_instance, shift)
    fragment = replace_instance(fragment, group_index, instance_index, tone_row_instance)
    return fragment


def apply_transposition(fragment: Fragment, max_transposition: int) -> Fragment:
    """
    Transpose one random series (transformed tone row instance).

    :param fragment:
        a fragment to be modified
    :param max_transposition:
        maximum interval of transposition (in semitones)
    :return:
        modified fragment
    """
    group_index, instance_index = draw_random_indices(
        fragment.mutable_sonic_content_indices, fragment.n_tone_row_instances_by_group
    )
    tone_row_instance = find_instance_by_indices(fragment, group_index, instance_index)
    shift = random.randint(-max_transposition, max_transposition)
    tone_row_instance = transpose_tone_row(tone_row_instance, shift)
    fragment = replace_instance(fragment, group_index, instance_index, tone_row_instance)
    return fragment


def create_transformations_registry(
        max_rotation: int, max_transposition: int
) -> TRANSFORMATION_REGISTRY_TYPE:
    """
    Get mapping from names to corresponding transformations and their arguments.

    :param max_rotation:
        maximum size of rotation (in elements)
    :param max_transposition:
        maximum interval of transposition (in semitones)
    :return:
        registry of transformations
    """
    registry = {
        'duration_change': (apply_duration_change, [get_duration_changes()]),
        'pause_shift': (apply_pause_shift, []),
        'inversion': (apply_inversion, []),
        'reversion': (apply_reversion, []),
        'rotation': (apply_rotation, [max_rotation]),
        'transposition': (apply_transposition, [max_transposition]),
    }
    return registry


def transform(
        fragment: Fragment,
        n_transformations: int,
        transformation_registry: TRANSFORMATION_REGISTRY_TYPE,
        transformation_names: list[str],
        transformation_probabilities: list[float]
) -> Fragment:
    """
    Apply multiple random transformations to a fragment.

    :param fragment:
        a fragment to be modified
    :param n_transformations:
        number of transformations to be applied
    :param transformation_registry:
        mapping from names to corresponding transformations and their arguments
    :param transformation_names:
        names of transformations to choose from
    :param transformation_probabilities:
        probabilities of corresponding transformations; this argument must have the same length
        as `transformation_names`
    :return:
        modified fragment
    """
    names_of_transformations_to_be_applied = random.choices(
        transformation_names,
        transformation_probabilities,
        k=n_transformations
    )
    for transformation_name in names_of_transformations_to_be_applied:
        transformation_fn, args = transformation_registry[transformation_name]
        fragment = transformation_fn(fragment, *args)
    fragment = override_calculated_attributes(fragment)
    return fragment
